{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "(dynamic_types)=\n",
    "# Динамическая типизация \n",
    "\n",
    "Следующий код `python` исполняется без ошибок."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0 <class 'int'>\n",
      "0.0 <class 'float'>\n",
      "zero <class 'str'>\n",
      "[0, 0.0, 'zero'] <class 'list'>\n"
     ]
    }
   ],
   "source": [
    "a = 0               # a - переменная типа int\n",
    "print(a, type(a))\n",
    "a = 0.0             # теперь a - переменная типа float\n",
    "print(a, type(a))\n",
    "a = \"zero\"          # теперь a - переменная типа str\n",
    "print(a, type(a))\n",
    "a = [0, 0.0, \"zero\"] # теперь a - переменная типа list\n",
    "print(a, type(a))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Кажется, что переменная `a` меняет свой тип 3 раза. Как это возможно? Прежде чем ответить на этот вопрос, освежим в памяти, как устроена типизация в `C/C++`, и разберемся, почему это изначально может смущать.\n",
    "```{note}\n",
    "Функция [type](https://docs.python.org/3/library/functions.html#type) спрашивает у объекта его тип. \n",
    "```\n",
    "## Статическая типизация в `C++`\n",
    "\n",
    "Чтобы разобрать, как устроена типизация в `C/C++`, рассмотрим следующий код на языке `C/C++`.\n",
    "\n",
    "```c++\n",
    "#include<iostream>\n",
    "\n",
    "#define FIRSTBYTE   1\n",
    "#define SECONDBYTE  256\n",
    "#define THIRDBYTE   256*256\n",
    "#define FOURTHBYTE  256*256*256\n",
    "\n",
    "\n",
    "int main(){\n",
    "\t// 1\n",
    "\tint a = 65*FIRSTBYTE + 66*SECONDBYTE + 67*THIRDBYTE + 68*FOURTHBYTE;\n",
    "\tstd::cout << a << std::endl;                // 1145258561\n",
    "\n",
    "\t// 2\n",
    "\tchar *b = reinterpret_cast<char*>(&a);\n",
    "\t// 3\n",
    "\tfor(size_t i=0; i<4; ++i)\n",
    "\t\tstd::cout << b[i];                      // ABCD\n",
    "\tstd::cout << std::endl;\n",
    "\n",
    "\t// 4\n",
    "\ta = 'A';\n",
    "\tstd::cout << a;                             // 65\n",
    "\n",
    "\treturn 0;\n",
    "}\n",
    "```\n",
    "Если его скомпилировать и запустить, то вы должны получить следующее в стандартном потоке вывода.\n",
    "```sh\n",
    "1145258561\n",
    "ABCD\n",
    "65\n",
    "```\n",
    "```{note}\n",
    "Приведенный выше код опирается на предположение, что в системе целочисленный тип `int` занимает 4 байта, что не гарантируется стандартами `C++`.  \n",
    "```\n",
    "\n",
    "Разберем по шагам, что происходит в этой программе. \n",
    "\n",
    "### // 1\n",
    "\n",
    "В начале объявляется целочисленная переменная `a` типа `int`, которая инициализируется значением $65 \\cdot 256^0 + 66 \\cdot 256^1 + 67 \\cdot 256^{2} + 68 \\cdot 256^{3} = 1145258561$. \n",
    "```c++\n",
    "// 1\n",
    "int a = 65*FIRSTBYTE + 66*SECONDBYTE + 67*THIRDBYTE + 68*FOURTHBYTE;\n",
    "std::cout << a << std::endl;                // 1145258561\n",
    "```\n",
    "В момент объявления выделяется достаточная для хранения переменной типа `int` (4 байта на большинстве платформ) область оперативной памяти, и переменная `a`  связывается с этой областью памяти. Во момент инициализации в эту область памяти записывается число 1145258561. Это число намерено задано в разложении по степеням числа $256$, чтобы явно выделить байты этого числа: 65, 66, 67 и 68 (см. [статью в википедии](https://ru.wikipedia.org/wiki/%D0%9F%D0%BE%D1%80%D1%8F%D0%B4%D0%BE%D0%BA_%D0%B1%D0%B0%D0%B9%D1%82%D0%BE%D0%B2) про порядок байтов, если непонятно о чем идёт речь).  \n",
    "### // 2\n",
    "\n",
    "Далее объявляется указатель `b` на символьную переменную типа `char`, который инициализируется адресом памяти переменной `a`. \n",
    "```c++\n",
    "// 2\n",
    "char *b = reinterpret_cast<char*>(&a);\n",
    "```\n",
    "Чтобы провернуть такой трюк, необходимо сначала поменять тип указателя, т.к. переменная `a` имеет 32-битный целочисленный тип `int`, а переменная `b` является указателем 8-битного типа `char` (который формально тоже является целочисленным). Компилятор запретил бы такую операцию из-за несоответствия типов (и невозможности их приведения друг к другу безопасным образом), поэтому в коде используется [reinterpret_cast](https://en.cppreference.com/w/cpp/language/reinterpret_cast), который вынуждает компилятор привести тип указателя.  \n",
    "\n",
    "\n",
    "```{warning}\n",
    "Такие возможности языка `C/C++` приводятся здесь только с целью демонстрации механизма работы переменных в `C/C++`. В настоящих программах рекомендуется избегать таких приёмов.\n",
    "```\n",
    "\n",
    "### // 3\n",
    "\n",
    "Далее в цикле выводятся значения `b[0]`, `b[1]`, `b[2]` и `b[3]`.\n",
    "```c++\n",
    "// 3\n",
    "for(size_t i=0; i<4; ++i)\n",
    "\tstd::cout << b[i];\n",
    "std::cout << std::endl;\n",
    "```\n",
    "В результате видим строку `ABCD`. Чтобы разобраться с тем, как это произошло, необходимо учесть \n",
    "- `b` --- указатель, который хранит в себе адрес переменной `a`;\n",
    "- однако `b` --- указатель 8-битного типа, а значит при его разыменовании (`b[0]`, что то же самое, что и `*b`) захватывается область размером всего в 1 байт.\n",
    "- `b[i]` эквивалентно `*(b + i) `, т.е. то же самое, что и `b[0]`, но на `i` байт правее.\n",
    "- т.к. в и тоге выводится символьная переменная, то она выводится согласно таблице [ASCII](https://ru.wikipedia.org/wiki/ASCII), в которой 65 соответствует символу \"`A`\", 66 --- символу \"`B`\", 67 --- символу \"`C`\", 68 --- символу \"`D`\".\n",
    "\n",
    "Следующий рисунок иллюстрирует примерное устройство памяти в этой программе на этот момент времени.\n",
    "```{figure} /_static/lecture_specific/dynamic_typing/cpp.png\n",
    "```\n",
    "### // 4\n",
    "\n",
    "Далее в переменную `a` записывается значение \"`'A'`\";\n",
    "```c++\n",
    "// 4\n",
    "a = 'A';\n",
    "std::cout << a; \n",
    "```\n",
    "Из-за того, что `C/C++` --- статически типизированный язык, то компилятор знает, что `a` --- переменная типа `int` (с таким типом она была объявлена), и `A` --- значение типа `char`, которое может быть без потерь приведено к `int` (8 бит меньше 32 бит). Исходя из этой информации, компилятор допускает эту операцию, неявно приводя символьное значение `'A'` к целочисленному значению `65` и записывая в соответствующую переменной `a` область памяти это значение.\n",
    "\n",
    "### Итого\n",
    "\n",
    "Итого, опуская технические подробности:\n",
    "- переменная соответствует области в памяти;\n",
    "- у переменной всегда есть определенный тип, он указывается при объявлении переменной и известен на этапе компиляции; \n",
    "- в самой области памяти нет информации о том, как правильно прочитать последовательность бит, которую она содержит; \n",
    "- при обращении к переменной, её значение получается в результате интерпретации последовательности бит в соответствующей области памяти на основании типа этой переменной;\n",
    "- технически возможно прочитать одну и ту же область памяти по-разному, обращаясь к ней переменными разных типов;\n",
    "- оператор присвоения \"`=`\" записывает значение справа от него в область памяти, именованную переменной слева от него с предварительной проверкой совместимости типов.\n",
    "\n",
    "## Динамическая типизация в `python`\n",
    "\n",
    "В `python` тип хранится не в переменной, а в самом объекте, а сама переменная всего лишь ссылается на объект (в `python` всё объекты), ничего не подозревая о типе этого самого объекта. На самом деле переменная в `CPython` представляет собой обертку над указателем `C` типа `void`, который всегда автоматически разыменовывается, а оператор присвоения \"`=`\" связывает переменную слева от него с объектом справа от него (изменяет значение указателя).\n",
    "\n",
    "```{note}\n",
    "В связи с этим, иногда термин **\"переменная\"** (**variable**) избегается в отношении `python` и заменяется терминами \"имя\" (**name**), \"ссылка\" (**reference**) или идентификатор.  \n",
    "```\n",
    "```{note}\n",
    "В `python` работать с адресами памяти напрямую невозможно. Единственное исключение составляет функция `id`, которая в некоторых реализациях возвращает адрес памяти (целое число), где хранится объект. Эта функция нужна только для того, чтобы всегда можно было выяснить, ссылаются ли два имени на один и тот же объект. Обратится по этому адресу средствами `python` невозможно. \n",
    "```\n",
    "\n",
    "Вернемся к примеру из начала раздела и объясним, как он работает."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "a = 0               # a - переменная типа int\n",
    "print(a, type(a))\n",
    "a = 0.              # теперь a - переменная типа float\n",
    "print(a, type(a))\n",
    "a = \"zero\"          # теперь a - переменная типа str\n",
    "print(a, type(a))\n",
    "a = [0, 0., \"zero\"] # теперь a - переменная типа list\n",
    "print(a, type(a))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### # 1\n",
    "\n",
    "Первая инструкция `a = 0`. \n",
    "1. Сначала вычисляется значение выражения справа: создается объект целочисленного типа `int` содержащий в себе значение `0`. \n",
    "2. Так как имени `a` до этого объявлено не было, то такое имя создаётся и связывается с объектом справа."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0 <class 'int'>\n"
     ]
    }
   ],
   "source": [
    "a = 0               # a - переменная типа int\n",
    "print(a, type(a))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{figure} /_static/lecture_specific/dynamic_typing/a_int.png\n",
    "```\n",
    "\n",
    "### # 2\n",
    "\n",
    "Вторая инструкция `a = 0.0`. \n",
    "1. Создаётся объект типа `float`, содержащий значение `0.0`. \n",
    "2. Так как имя `a` уже существует, то оно связывается с новым объектом.  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "0.0 <class 'float'>\n"
     ]
    }
   ],
   "source": [
    "a = 0.0            # теперь a - переменная типа float\n",
    "print(a, type(a))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{figure} /_static/lecture_specific/dynamic_typing/a_float.png\n",
    "```\n",
    "\n",
    "### # 3\n",
    "\n",
    "Третья инструкция `a=\"zero\"`. \n",
    "1. Создаётся объект строкового типа `str`, содержащий значение `\"zero\"`. \n",
    "2. Существующее имя `a` связывается с новым объектом, который имеет другой тип. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "zero <class 'str'>\n"
     ]
    }
   ],
   "source": [
    "a = \"zero\"          # теперь a - переменная типа str\n",
    "print(a, type(a))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```{figure} /_static/lecture_specific/dynamic_typing/a_str.png\n",
    "```\n",
    "### # 4\n",
    "\n",
    "Четвертая инструкция `a = [0, 0.0, \"zero\"]`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[0, 0.0, 'zero'] <class 'list'>\n"
     ]
    }
   ],
   "source": [
    "a = [0, 0.0, \"zero\"] # теперь a - переменная типа list\n",
    "print(a, type(a))"
   ]
  },
  {
   "attachments": {},
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "\n",
    "Попытайтесь сами объяснить, что происходит при её выполнении, а затем сверьтесь со спойлером ниже. \n",
    "\n",
    "\n",
    "````{toggle} Объяснение\n",
    "1. Создаётся три объекта: целочисленный объект `0`, действительнозначный объект `0.0` и строковый объект `\"zero\"`.\n",
    "2. Создаётся объект спискового объект длинны 3, ячейки которого связываются с объектами, созданными на предыдущем этапе.\n",
    "3. Существующее имя `a` связывается связывается с новым объектом, который имеет другой тип. \n",
    "```{figure} /_static/lecture_specific/dynamic_typing/a_list.png\n",
    "```\n",
    "````\n",
    "\n",
    "\n",
    "```{note}\n",
    "  Иногда говорят, что тип объекта \"живёт\" в самом объекте, а не в \"переменной\". \n",
    "```\n",
    "\n",
    "### Итого\n",
    "\n",
    "Итого, опуская технические подробности:\n",
    "- вместо переменных в `python` имена;\n",
    "- у имени нет типа; \n",
    "- имя всегда ссылается на какой-то объект (не бывает неинициализированных имён), внутри которого кроме полезных данных хранится служебная информация о типе объекта и т.п.;\n",
    "- оператор присвоения \"`=`\" связывает имя слева от него с объектом с права от него: если такого имени ещё нет, то оно создаётся, если такое имя уже существует (связанно с каким-то объектом), то ссылка перекидывается на новый объект."
   ]
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "eea3ef54c04064aa17c056cc2578c78db8b44278034b77b7225a3166c34cea02"
  },
  "kernelspec": {
   "display_name": "Python 3.8.10 ('venv': venv)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.7 (tags/v3.10.7:6cc6b13, Sep  5 2022, 14:08:36) [MSC v.1933 64 bit (AMD64)]"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "2cf0fd39df186aadf3d4a87586a022f968cfd977f53ee0e019844353ece0ae6e"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
